;; load balancing example

;; parameters

(defparameter *state* #p"inputs/revivals-l-11-n-11/")

(defparameter *j* 1/25) ;; *u* = 1

(defparameter *maxdim* '(nil nil 100))

;; libraries

(load "conf.lisp")
(require :tebdol)

(use-package '(:blas :tensor :mpi :part :mps :tebd :bhm))
(setf *print-level* 2)

;; hamiltonian

(defun make-hamiltonian (dimension length j)
  (loop
     with h = (bose-hubbard-site-hamiltonian dimension j 1 0)
     with m = (make-array length)
     for i from 1 below (1- length)
     do
       (setf (svref m i) h)
     finally
       (setf (svref m 0) (tensor-contract (left-open-boundary-tensor h) 1 h 0))
       (setf (svref m (1- length)) (tensor-contract h 3 (right-open-boundary-tensor h) 0))
       (return m)))

;;
;; bose-hubbard quasimomentum
;;
;;      -1    0 0    1
;;
;;         +------+
;; -1      | b+ 0 |  1
;;    +----+------+----+
;;  0 |  0 |  1 0 |  0 |
;;  0 | b+ |  n 1 | b- |
;;    +----+------+----+
;;  1    1 | b- 0 |
;;         +------+
;;

(defun site-quasimomentum (dimension q i)
  (let ((ops (list
	      (cons  '0 #'(lambda (n) (declare (ignore n)) 0))
	      (cons  '1 #'(lambda (n) (declare (ignore n)) 1))
	      (cons 'b- #'(lambda (n) (* (exp (* +blas-i+ q i)) (sqrt n))))
	      (cons 'b+ #'(lambda (n) (* (exp (* -1 +blas-i+ q i)) (sqrt (1+ n)))))
	      (cons  'n #'(lambda (n) n))))
	(tbl (flet ((f (x y z)
		      (cons (list (make-subscript
				   :numbers (list (first x))
				   :subscript (second x))
				  (make-subscript
				   :numbers (list (first y))
				   :subscript (second y)))
			    z)))
	       (list
		(f '(-1 0) '( 0 0) 'b+)
		(f '(-1 0) '( 1 0) '1)
		(f '( 0 0) '( 0 0) '1)
		(f '( 0 1) '(-1 0) 'b+)
		(f '( 0 1) '( 0 0) 'n)
		(f '( 0 1) '( 0 1) '1)
		(f '( 0 1) '( 1 0) 'b-)
		(f '( 1 0) '(-1 0) '1)
		(f '( 1 0) '( 0 0) 'b-)))))
    (let ((l (list (make-segment :numbers '(-1) :dimension 1)
		   (make-segment :numbers '(0) :dimension 2)
		   (make-segment :numbers '(1) :dimension 1))))
      (functional-tensor
       (list l (ket-physical-index dimension) (bra-physical-index dimension) l)
       #'(lambda (s)
	(let ((x (find
		  (list (first s) (fourth s))
		  tbl
		  :test #'equalp
		  :key #'car)))
	  (if x
	      (funcall (cdr (find (cdr x) ops :key #'car))
		       (- (first (subscript-numbers (third s)))))
	      0)))))))

(defun make-quasimomentum-mpo (dimension length q)
  (loop
     with m = (make-array length)
     for i from 1 below (1- length)
     do
       (setf (svref m i) (site-quasimomentum dimension q (+ i (- 1/2 (/ length 2)))))
     finally
       (let ((o (site-quasimomentum dimension q (- 1/2 (/ length 2)))))
	 (setf (svref m 0) (tensor-contract (left-open-boundary-tensor o) 1 o 0)))
       (let ((o (site-quasimomentum dimension q (- (/ length 2) 1/2))))
	 (setf (svref m (1- length)) (tensor-contract o 3 (right-open-boundary-tensor o) 0)))
       (return m)))

;; initialization

(mpi-init)

(defparameter *rank* (mpi-comm-rank *mpi-comm-world*))
(defparameter *size* (mpi-comm-size *mpi-comm-world*))

(defparameter *length* (state-parameter *state* :model-length))
(defparameter *particles* (state-parameter *state* :particle-number))
(defparameter *dimension* (state-parameter *state* :site-dimension))

(defparameter *partition* (uniform-partition *size* *length*))

(defparameter *mps* (load-mps *rank* *partition* *state*))
(defparameter *mph* (make-hamiltonian *dimension* *length* *j*))
(defparameter *mpq* (make-quasimomentum-mpo *dimension* *length* 0))

(defparameter *count* 100)
(defparameter *step* (/ (* 5 pi) *count*))
(defparameter *propagator* (make-mph-propagator *mph* *step*))

(when (zerop *rank*)
  (format
   t
   "# processes = ~a; length = ~a; particles = ~a; maxdim = ~a; j = ~f;~%"
   *size* *length* *particles* *maxdim* *j*)
  (format t "# time n(quasimomentum=0;time) norm~%"))

(mps-prepare *rank* *length* *mps* *maxdim*)

;; main loop

(loop for i from 0 to *count* do
     (let ((q (realpart (mps-expectation-value *size* *rank* *length* *mpq* *mps*)))
	   (n (realpart (mps-overlap *size* *rank* *length* *mps* *mps*))))
       (if (= *rank* (1- *size*))
	   (format t "~,2f ~f ~f~%"
		   (/ (* i *step*) pi)
		   (/ q n *length*)
		   (sqrt n))))
     (when (< i *count*)
       (let ((subwt (make-array (submps-size *mps*) :initial-element 0)))
	 (tebd-evolve *rank* *length* *propagator* *mps* *maxdim* :subwt subwt)
	 (let ((p (balance-partition *size* *rank* (coerce subwt 'list))))
	   (when (zerop *rank*)
	     (format *error-output* "new partition = ~a~%" p))
	   (setf *mps* (change-partition *size* *rank* *length* *mps* *partition* p))
	   (setf *partition* p)))))

(mpi-finalize)
