;; hard-core boson expansion in a two-dimensional optical lattice

;; parameters

(defparameter *width* 6)
(defparameter *height* 6)
(defparameter *diameter* 2)

(defparameter *jy* 5d-1) ;; *jx* = 1

(defparameter *step* (/ 16d0))

(defparameter *maxdim*
  (list (* (expt 2 32) double-float-epsilon) nil 128))

;; libraries

(load "conf.lisp")
(require :tebdol)

(use-package '(:blas :ttns :tebd :bhm))
(setf *print-level* 2)

;; functions

(defun radius-x (pn n)
  (loop
     with w = (array-dimension pn 0)
     with h = (array-dimension pn 1)
     with c = (- (/ w 2) 5d-1)
     for i below w
     sum (* (expt (- i c) 2)
	    (loop
	       for j below h
	       sum (aref pn i j)))
     into s
     finally (return (sqrt (/ s n)))))

(defun radius-y (pn n)
  (loop
     with w = (array-dimension pn 0)
     with h = (array-dimension pn 1)
     with c = (- (/ h 2) 5d-1)
     for j below h
     sum (* (expt (- j c) 2)
            (loop
	       for i below w
	       sum (aref pn i j)))
     into s
     finally (return (sqrt (/ s n)))))

(defun make-square-ttns (width height diameter dimension)
  (loop
     with p = (make-vacuum-ttns width height dimension)
     with xshift = (ceiling (- width diameter) 2)
     with yshift = (floor (- height diameter) 2)
     for i below diameter
     do
       (loop
	  for j below diameter
	  do (ttns-create-particle p (+ i xshift) (+ j yshift) dimension))
     finally
       (compress-ttns p nil)
       (return p)))

;; main

;; hard-core bosons
(defparameter *dimension* 2)
(defparameter *particles* (expt *diameter* 2))
(defparameter *potential* (make-array (list *width* *height*) :initial-element 0d0))

(defparameter *tbhsh* (bose-hubbard-site-hamiltonian *dimension* *jy* 0d0 0d0))
(defparameter *ssps* (make-ssps *width* *height* *dimension* 0d0 *potential* (/ *step* 2)))
(defparameter *dsps* (make-dsps *width* *height* *dimension* 0d0 *potential* (/ *step* 2)))
(defparameter *dspy* (make-double-site-propagator *tbhsh* *tbhsh* (/ *step* 2)))

(defparameter *ttns* (make-square-ttns *width* *height* *diameter* *dimension*))
(defparameter *pn* (all-pn-operators *width* *height* *dimension*))

(format t "# width = ~A~%" *width*)
(format t "# height = ~A~%" *height*)
(format t "# diameter = ~A~%" *diameter*)
(format t "# jy = ~A~%" *jy*)
(format t "# step = ~A~%" *step*)
(format t "# maxdim = ~A~%" *maxdim*)
(format t "~%# time radius-y radius-y imbalance norm~%# ----~%# density~%# ----~%~%")

(loop
   for i to 32
   do
     (unless (zerop i)
       (ttns-tebd-evolve *ttns* *ssps* *dsps* *dspy* *maxdim*))
     (let* ((n (realpart (ttns-overlap *ttns* *ttns*)))
	    (p (all-pn *ttns* n *pn*)))
       (format
	t
	"~,4F ~F ~F ~F~%"
	(* i *step*)
	(radius-x p *particles*)
	(radius-y p *particles*)
	n)
       (format t "----~%")
       (print-pn p)
       (format t "----~%")))
