;; many-body localization in a two-dimensional optical lattice

;; parameters

(defparameter *potential* #p"inputs/mbl-w-4-h-4.pot")
(defparameter *diameter* 2)
(defparameter *u* 1.6d1) ;; *j* = 1

(defparameter *dimension* 7)
(defparameter *step* (/ 16d0))

(defparameter *maxdim*
  (list (* (expt 2 32) double-float-epsilon) nil 128))

;; libraries

(load "conf.lisp")
(require :tebdol)

(use-package '(:blas :ttns :tebd :bhm))
(setf *print-level* 2)

;; functions

(defun make-half-circle-ttns (width height diameter dimension)
  (loop
     with p = (make-vacuum-ttns width height dimension)
     with xshift = (ceiling (- width diameter) 2)
     with yshift = (floor (- height diameter) 2)
     for i below (ceiling diameter 2)
     do
       (loop
	  for j below diameter
	  when (and (<=
		     (sqrt
		      (+ (expt (- i (/ (1- diameter) 2)) 2)
			 (expt (- j (/ (1- diameter) 2)) 2)))
		     (/ diameter 2))
		    (>= (+ i xshift) 0)
		    (>= (+ j yshift) 0)
		    (< (+ i xshift) width)
		    (< (+ j yshift) height))
	  do
	    (ttns-create-particle p (+ i xshift) (+ j yshift) dimension))
     finally
       (compress-ttns p nil)
       (return p)))

(defun imbalance (pn)
  (loop
     with w = (array-dimension pn 0)
     with h = (array-dimension pn 1)
     for i below w
     for s = (loop
		for j below h
		sum (aref pn i j))
     when (< i (/ w 2))
     sum s into l
     else
     sum s into r
     finally
       (return (/ (- l r) (+ l r)))))

(defun load-potential (pathname)
  (with-standard-io-syntax
    (with-open-file (s pathname)
      (read s))))

;; initialization

(defparameter *potarray* (load-potential *potential*))
(defparameter *width* (array-dimension *potarray* 0))
(defparameter *height* (array-dimension *potarray* 1))

(defparameter *tbhsh* (bose-hubbard-site-hamiltonian *dimension* 1d0 0d0 0d0))
(defparameter *ssps* (make-ssps *width* *height* *dimension* *u* *potarray* (/ *step* 2)))
(defparameter *dsps* (make-dsps *width* *height* *dimension* *u* *potarray* (/ *step* 2)))
(defparameter *dspy* (make-double-site-propagator *tbhsh* *tbhsh* (/ *step* 2)))

(defparameter *ttns* (make-half-circle-ttns *width* *height* *diameter* *dimension*))
(defparameter *pn* (all-pn-operators *width* *height* *dimension*))

(format t "# potential = ~A~%" *potential*)
(format t "# width = ~A~%" *width*)
(format t "# height = ~A~%" *height*)
(format t "# diameter = ~A~%" *diameter*)
(format t "# u = ~A~%" *u*)
(format t "# step = ~A~%" *step*)
(format t "# maxdim = ~A~%" *maxdim*)
(format t "~%# time imbalance norm~%# ----~%# density~%# ----~%~%")

;; main loop

(loop
   for i to 128
   do
     (unless (zerop i)
       (ttns-tebd-evolve *ttns* *ssps* *dsps* *dspy* *maxdim*))
     (let* ((n (realpart (ttns-overlap *ttns* *ttns*)))
	    (p (all-pn *ttns* n *pn*)))
       (format t "~,4F ~F ~F~%" (* i *step*) (imbalance p) n)
       (format t "----~%")
       (print-pn p)
       (format t "----~%")))
