;; equilibration in an optical lattice

;; parameters

(defparameter *state* #p"inputs/trotzky-l-21-n-7/")

(defparameter *j* 4.1d-1)
(defparameter *k* 4.1d-3)

(defparameter *maxdim* '(nil nil 100))

;; libraries

(load "conf.lisp")
(require :tebdol)

(use-package '(:tensor :mpi :part :mps :tebd :bhm))
(setf *print-level* 2)

;; hamiltonian

(defun make-hamiltonian (dimension length j k)
  (loop
     with c = (/ (1- length) 2)
     with m = (make-array length)
     for i from (1+ (- c)) below 0
     do
       (let ((o (bose-hubbard-site-hamiltonian dimension j 1 (* 1/2 k i i))))
	 (setf (svref m (+ c i)) o
	       (svref m (- c i)) o))
     finally
       (setf (svref m c) (bose-hubbard-site-hamiltonian dimension j 1 (* 1/2 k 0 0)))
       (let ((o (bose-hubbard-site-hamiltonian dimension j 1 (* 1/2 k c c))))
	 (setf (svref m 0) (tensor-contract (left-open-boundary-tensor o) 1 o 0))
	 (setf (svref m (1- length)) (tensor-contract o 3 (right-open-boundary-tensor o) 0)))
       (return m)))

;; odd particle number operator
;;
;;       0 0
;;    +------+
;;  0 |  1 0 |
;;  0 |  n 1 |
;;    +------+
;;
;;       0 0
;;    +------+
;;  0 |  1 0 |
;;  0 |  0 1 |
;;    +------+

(defun site-operator (dimension type)
  (let ((ops (list
	      (cons '1 #'(lambda (n) (declare (ignore n)) 1))
	      (cons 'n #'(lambda (n) n))))
	(tbl (flet ((f (x y z)
		      (cons (list (make-subscript
				   :numbers (list (first x))
				   :subscript (second x))
				  (make-subscript
				   :numbers (list (first y))
				   :subscript (second y)))
			    z)))
	       (case type
		 (:identity
		  (list
		   (f '(0 0) '(0 0) '1)
		   (f '(0 1) '(0 1) '1)))
		 (:particle-number
		  (list
		   (f '(0 0) '(0 0) '1)
		   (f '(0 1) '(0 0) 'n)
		   (f '(0 1) '(0 1) '1)))))))
    (let ((l (list (make-segment :numbers '(0) :dimension 2))))
      (functional-tensor
       (list l (ket-physical-index dimension) (bra-physical-index dimension) l)
       #'(lambda (s)
	(let ((x (find
		  (list (first s) (fourth s))
		  tbl
		  :test #'equalp
		  :key #'car)))
	  (if x
	      (funcall (cdr (find (cdr x) ops :key #'car))
		       (- (first (subscript-numbers (third s)))))
	      0)))))))

(defun make-particle-number-mpo (dimension length function)
  (loop
     with c = (/ (1- length) 2)
     with m = (make-array length)
     with e = (site-operator dimension :identity)
     with p = (site-operator dimension :particle-number)
     for i from (1+ (- c)) to (1- c)
     do
       (setf (svref m (+ i c))
	     (if (funcall function i) p e))
     finally
       (let ((o (if (funcall function (- c)) p e)))
	 (setf (svref m 0) (tensor-contract (left-open-boundary-tensor o) 1 o 0)))
       (let ((o (if (funcall function c) p e)))
	 (setf (svref m (1- length)) (tensor-contract o 3 (right-open-boundary-tensor o) 0)))
       (return m)))

;; initialization

(mpi-init)

(defparameter *rank* (mpi-comm-rank *mpi-comm-world*))
(defparameter *size* (mpi-comm-size *mpi-comm-world*))

(defparameter *length* (state-parameter *state* :model-length))
(defparameter *particles* (state-parameter *state* :particle-number))
(defparameter *dimension* (state-parameter *state* :site-dimension))

(defparameter *partition* (uniform-partition *size* *length*))

(defparameter *mps* (load-mps *rank* *partition* *state*))
(defparameter *mph* (make-hamiltonian *dimension* *length* *j* *k*))
(defparameter *mpn* (make-particle-number-mpo *dimension* *length* #'oddp))

(defparameter *count* 100)
(defparameter *step* (/ (* 5 (/ pi 2 *j*)) *count*))
(defparameter *propagator* (make-mph-propagator *mph* *step*))

(when (zerop *rank*)
  (format
   t
   "# processes = ~a; length = ~a; particles = ~a; maxdim = ~a; j = ~f; k = ~f;~%"
   *size* *length* *particles* *maxdim* *j* *k*)
  (format t "# time nodd norm~%"))

(mps-prepare *rank* *length* *mps* *maxdim*)

;; main loop

(loop for i from 0 to *count* do
     (let ((e (realpart (mps-expectation-value *size* *rank* *length* *mpn* *mps*)))
	   (n (realpart (mps-overlap *size* *rank* *length* *mps* *mps*))))
       (if (= *rank* (1- *size*))
	   (format t "~,2f ~f ~f~%"
		   (/ (* i *step*) (/ pi 2 *j*))
		   (/ e n *particles*)
		   (sqrt n))))
     (when (< i *count*)
       (tebd-evolve *rank* *length* *propagator* *mps* *maxdim*)))

(mpi-finalize)
