;; random lattice potential

;; parameters

(defparameter *width* 4)
(defparameter *height* 4)
(defparameter *diameter* 2)
(defparameter *epp* 1.4d-1)
(defparameter *fwhm* 1d0)

;; fixed seed
(defparameter *seed* 0)

;; random seed, uncomment to get a real random potential
;; (defparameter *seed* (random #x100000000 (make-random-state t)))

;; libraries

(load (merge-pathnames "gsl.lisp" *load-truename*))
(load (merge-pathnames "conv.lisp" *load-truename*))
(use-package '(:gsl :conv))

;; main

(defun harmonic-potential (width height)
  (loop
     with a = (make-array (list width height))
     for i below width
     do (loop
	   for j below height
	   do (setf (aref a i j)
		    (+ (expt (- i (/ (1- width) 2)) 2)
		       (expt (- j (/ (1- height) 2)) 2))))
     finally (return a)))

(defun half-circle-mean-energy (potential diameter)
  (loop
     with m = (ceiling (- (array-dimension potential 0) diameter) 2)
     with n = (floor (- (array-dimension potential 1) diameter) 2)
     with p = 0
     with e = 0
     for i below (/ diameter 2)
     do (loop
	   for j below diameter
	   when (<= (sqrt (+ (expt (- i (/ (1- diameter) 2)) 2)
			     (expt (- j (/ (1- diameter) 2)) 2)))
		    (/ diameter 2))
	   do
	     (incf p)
	     (incf e (aref potential (+ i m) (+ j n))))
     finally (return (/ e p))))

(defun scaled-harmonic-potential (width height diameter epp)
  (loop
     with a = (harmonic-potential width height)
     with c = (/ (* 24.4d0 epp) (half-circle-mean-energy a diameter))
     for i below (array-total-size a)
     do (setf (row-major-aref a i) (* c (row-major-aref a i)))
     finally (return a)))

(defparameter *gsl-rng* (gsl-rng-alloc +gsl-rng-ranlxd2+))

(gsl-rng-set *gsl-rng* *seed*)

(defun binomial-variate (p n)
  (gsl-ran-binomial *gsl-rng* p n))

(defun squared-binomial-mean (p n)
  (/ (+ p (* (expt p 2) (1- n))) n))

(defconstant +squared-binomial-fwhm+ 1.06d-1) ;; measured value

(defun squared-binomial-potential (width height fwhm)
  (loop
     with c = (/ fwhm +squared-binomial-fwhm+)
     with m = (squared-binomial-mean 5d-1 49)
     with a = (make-array (list width height))
     for i below (array-total-size a)
     do (setf (row-major-aref a i)
	      (* c (- (expt (/ (binomial-variate 5d-1 49) 49) 2) m)))
     finally (return a)))

(defun convolved-potential (width height fwhm)
  (loop
     with p = (squared-binomial-potential (+ width 8) (+ height 8) fwhm)
     with g = (normalized-gaussian 9 9 5d-1)
     with a = (convolve p g)
     with b = (make-array (list width height))
     for i below width
     do (loop
	   for j below height
	   do (setf (aref b i j)
		    (aref a (+ i 4) (+ j 4))))
     finally (return b)))

(defun full-potential (width height diameter epp fwhm)
  (loop
     with a = (scaled-harmonic-potential width height diameter epp)
     with b = (convolved-potential width height fwhm)
     for i below (array-total-size a)
     do (incf (row-major-aref a i) (row-major-aref b i))
     finally (return a)))

(defun save-potential (potential pathname)
  (when (or (not (probe-file pathname)) (y-or-n-p "Overwrite potential ~S?" pathname))
    (with-standard-io-syntax
      (let ((*print-pretty* t))
	(with-open-file (s pathname :direction :output :if-exists :supersede)
	  (format s ";; width = ~A, height = ~A, diameter = ~A, epp = ~A, fwhm = ~A, seed = ~A~%"
		  *width* *height* *diameter* *epp* *fwhm* *seed*)
	  (prin1 potential s)
	  (terpri s))))))

(save-potential
 (full-potential *width* *height* *diameter* *epp* *fwhm*)
 (pathname
  (format nil "inputs/mbl-w-~A-h-~A.pot" *width* *height*)))

(gsl-rng-free *gsl-rng*)
