;; parameters

(defparameter *dimension* 15) ;; physical dimension
(defparameter *length* 11)
(defparameter *particles* 11)
(defparameter *maxdim* '(nil nil 100))

;; libraries

(load "conf.lisp")
(require :tebdol)

(use-package '(:array :tensor :mps :bhm))
(setf *print-level* 2)

;; main

(defun left-normalize (mps maxdim)
  (let ((n (length mps)))
    (loop for i below (1- n) do
	 (multiple-value-bind (l s r) (tensor-decompose (svref mps i) '(0 1) maxdim :left)
	   (declare (ignore s))
	   (setf (svref mps i) l)
	   (setf (svref mps (1+ i)) (tensor-contract r 1 (svref mps (1+ i)) 0))))
    (setf (svref mps (1- n)) (tensor-normalize (svref mps (1- n))))))

(defun right-compress (mps maxdim)
  (let ((n (length mps)))
    (loop for i from (1- n) above 0 do
         (multiple-value-bind (l s r) (tensor-decompose (svref mps i) '(0) maxdim :right)
	   (declare (ignore s))
           (setf (svref mps i) r)
           (setf (svref mps (1- i)) (tensor-contract (svref mps (1- i)) 2 l 0))))))

;;
;; particle in a box creation operator
;;
;;      -1   0
;;
;;    +----+---+
;;  0 | b+ | 1 |
;;    +----+---+
;;  1 |  1 | 0 |
;;    +----+---+
;;

(defun site-piab (dimension weight)
  (let ((ops (list
	      (cons  '1 #'(lambda (n) (declare (ignore n)) 1))
	      (cons 'b+ #'(lambda (n) (* weight (sqrt (1+ n)))))))
	(tbl (flet ((f (x y z)
		      (cons (list (make-subscript
				   :numbers (list (first x))
				   :subscript (second x))
				  (make-subscript
				   :numbers (list (first y))
				   :subscript (second y)))
			    z)))
	       (list
		(f '(0 0) '(-1 0) 'b+)
		(f '(0 0) '( 0 0) '1)
		(f '(1 0) '(-1 0) '1)))))
    (functional-tensor
     (list
      (list (make-segment :numbers '(0) :dimension 1)
	    (make-segment :numbers '(1) :dimension 1))
      (ket-physical-index dimension)
      (bra-physical-index dimension)
      (list (make-segment :numbers '(-1) :dimension 1)
	    (make-segment :numbers '(0) :dimension 1)))
     #'(lambda (s)
      (let ((x (find
		(list (first s) (fourth s))
		tbl
		:test #'equalp
		:key #'car)))
	(if x
	    (funcall (cdr (find (cdr x) ops :key #'car))
		     (- (first (subscript-numbers (third s)))))
	    0))))))

(defun right-tensor (operator)
  (let* ((i (fourth (tensor-indices operator))))
    (functional-tensor
     (list (conjugate-index i)
	   (list (make-segment :numbers '(-1) :dimension 1)))
     #'(lambda (s)
      (if (zerop (subscript-subscript (first s)))
	  1
	  0)))))

(defun make-piab-mpo (dimension length)
  (loop
     with m = (make-array length)
     for i from 1 below (1- length)
     do
       (setf (svref m i) (site-piab dimension (sin (/ (* pi (1+ i)) (1+ length)))))
     finally
       (let ((o (site-piab dimension (sin (/ (* pi 1) (1+ length))))))
	 (setf (svref m 0) (tensor-contract (left-open-boundary-tensor o) 1 o 0)))
       (let ((o (site-piab dimension (sin (/ (* pi length) (1+ length))))))
	 (setf (svref m (1- length)) (tensor-contract o 3 (right-tensor o) 0)))
       (return m)))

(defun apply-and-fuse (mpo mps)
  (let* ((n (length mpo))
	 (o (make-array n)))
    (dotimes (i n o)
      (setf
       (svref o i)
       (tensor-fuse
	(tensor-permute
	 (tensor-contract (svref mpo i) 2 (svref mps i) 1)
	 '(0 2 3 1 4))
	'(2 3)
	'(:normal :normal :reverse))))))

(defun make-vacuum-mps (dimension length)
  (let* ((m (make-array length))
	 (l (list (make-segment :numbers '(0) :dimension 1)))
	 (u (functional-tensor
	     (list l (ket-physical-index dimension) l)
	     #'(lambda (s) (declare (ignore s)) 1))))
    (dotimes (i length m)
      (setf (svref m i) u))))

(defparameter *mps* (make-vacuum-mps *dimension* *length*))
(defparameter *mpp* (make-piab-mpo *dimension* *length*))

(dotimes (i *particles*)
  (format t "creating particle #~a~%" (1+ i))
  (setf *mps* (apply-and-fuse *mpp* *mps*))
  (left-normalize *mps* *maxdim*)
  (right-compress *mps* *maxdim*))

(save-mps
 *mps*
 *dimension*
 *particles*
 (make-pathname
  :directory
  (list :relative (format nil "inputs/revivals-l-~A-n-~A" *length* *particles*))))
