(defpackage :array
  (:use :common-lisp :sb-sys :sb-ext :sb-alien :util :blas)
  (:export :array-copy
	   :array-scalar-/
	   :array-conjugate
	   :array-norm
	   :array-permute
	   :array-contract
	   :array-decompose
	   :array-addf
	   :*svd-driver*))

(in-package :array)

(defun array-copy (array)
  (let ((new-array (make-blas-array (array-dimensions array))))
    (dotimes (i (array-total-size array) new-array)
      (setf (row-major-aref new-array i) (row-major-aref array i)))))

(defmacro define-unary-element-wise-operation (name function)
  `(defun ,name (array &rest args)
     (let ((new-array (make-blas-array (array-dimensions array))))
       (dotimes (i (array-total-size array) new-array)
	 (setf (row-major-aref new-array i)
	       (apply (function ,function)
		      (row-major-aref array i)
		      args))))))

(define-unary-element-wise-operation array-scalar-/ /)
(define-unary-element-wise-operation array-conjugate conjugate)

(defun array-norm (array)
  (let ((acc 0))
    (dotimes (i (array-total-size array) (sqrt acc))
      (incf acc (+ (expt (realpart (row-major-aref array i)) 2)
                   (expt (imagpart (row-major-aref array i)) 2))))))

;; compiler can produce optimized code for array operations

;; array permutation generator

(defmacro array-permute-macro (array permutation)
  (let* ((a (gensym))
	 (b (gensym))
	 (rank (length permutation))
	 (mi (loop for i below rank collect (gensym)))
	 (md (loop for i below rank collect (gensym)))
	 (mip (list-permute mi permutation))
	 (mdp (list-permute md permutation)))
    (labels ((nest (&optional (i 0))
    	       (if (eql i rank)
    		   `(setf (aref ,b ,@mip) (aref ,a ,@mi))
		   `(dotimes (,(nth i mi) ,(nth i md)
			      ,@(when (eql i 0) `(,b)))
		      ,(nest (1+ i))))))
      `(let* ((,a ,array)
	      ,@(loop for i below rank
		   collect `(,(nth i md) (array-dimension ,a ,i)))
	      (,b (make-array (list ,@mdp) :element-type 'blas-float)))
	 (declare (type (simple-array blas-float) ,a ,b))
	 ,(nest)))))

;; generate array permutation functions dynamically

(let ((cache (make-hash-table :test 'equal)))
  (defun array-permute (array permutation)
    (funcall (or (gethash permutation cache)
		 (setf (gethash permutation cache)
		       (symbol-function
			(compile (intern (format nil "ARRAY-PERMUTE-~{~D~^-~}" permutation))
				 `(lambda (a) (array-permute-macro a ,permutation))))))
	     array)))

;;; contraction function generator

(defun make-array-contract-fn (ra ia rb ib)
  (labels ((perm (a rx ix s)
	     `(setf ,a (array-permute-macro ,a ,(make-indices-permutation rx ix s))))
	   (syms (rx ix)
	     (let* ((d (loop for i below rx for g = (gensym) collect g))
		    (k (loop for i in ix collect (nth i d)))
		    (l (loop for i below rx unless (member i ix) collect (nth i d))))
	       (values d k l)))
	   (alien (a)
	     `(sap-alien (vector-sap (array-storage-vector ,a)) (* double)))
	   (contract (pa pb)
	     (let (transa transb lda ldb)
	       (if (eql pa :left)
		   (setf transb "t" ldb 'n)
		   (setf transb "n" ldb 'k))
	       (if (eql pb :left)
		   (setf transa "n" lda 'm)
		   (setf transa "t" lda 'k))
	       `(with-pinned-objects
		    ((array-storage-vector a)
		     (array-storage-vector b)
		     (array-storage-vector c)
		     (array-storage-vector *blas-alien-0*)
		     (array-storage-vector *blas-alien-1*))
		  (zgemm ,transa ,transb m n k ,(alien *blas-alien-1*) ,(alien 'b) ,lda
			 ,(alien 'a) ,ldb ,(alien '*blas-alien-0*) ,(alien 'c) m)
		  c))))

    (multiple-value-bind (da ka la) (syms ra ia)
      (multiple-value-bind (db kb lb) (syms rb ib)
	`(lambda (a b)
	   (declare (type (simple-array blas-float) a b))
	   (destructuring-bind (,@da) (array-dimensions a)
	     (destructuring-bind (,@db) (array-dimensions b)
	       (unless (and ,@(loop for i in ka for j in kb collect `(= ,i ,j)))
		 (error "Array dimensions do not match."))
	       (let ((k (* ,@ka))
		     (n (* ,@la))
		     (m (* ,@lb))
		     (c (make-array (list ,@la ,@lb) :element-type 'blas-float)))
		 (declare (type (simple-array blas-float) c))
		 ,@(let ((pa (indices-position ra ia))
			 (pb (indices-position rb ib))
			 (sa (indices-shape ia))
			 (sb (indices-shape ib)))
			(cond ((and pa pb)
			       (if (equal sa sb)
				   `(,(contract pa pb))
				   `((if (< (array-total-size a) (array-total-size b))
					 (progn
					   ,(perm 'a ra ia sb)
					   ,(contract :left pb))
					 (progn
					   ,(perm 'b rb ib sa)
					   ,(contract pa :left))))))
			      ((and pa (not pb))
			       `(,(perm 'b rb ib sa) ,(contract pa :left)))
			      ((and (not pa) pb)
			       `(,(perm 'a ra ia sb) ,(contract :left pb)))
			      (t
			       (let ((s (loop for i below (length ia) collect i)))
				 `(,(perm 'a ra ia s)
				   ,(perm 'b rb ib s)
				   ,(contract :left :left))))))))))))))

;;; generate array contraction functions dynamically

(let ((cache (make-hash-table :test 'equal)))
  (defun array-contract (a ia b ib)
    (unless (listp ia)
      (setf ia (list ia)))
    (unless (listp ib)
      (setf ib (list ib)))
    (let ((key (list (array-rank a) ia (array-rank b) ib)))
      (funcall (or (gethash key cache)
		   (setf (gethash key cache)
			 (symbol-function
			  (compile
			   (intern
			    (format nil "ARRAY-CONTRACT-~{R~D-~{~D~^-~}-R~D-~{~D~^-~}~}" key))
			   (apply #'make-array-contract-fn key)))))
	       a
	       b))))

;; array decomposition

(defun array-decompose-zgesvd (m n a s u vt)
  (let* ((min (min m n))
	 (rwork (make-double-array (* 5 min))))
    (with-pinned-objects
	((array-storage-vector a)
	 (array-storage-vector s)
	 (array-storage-vector u)
	 (array-storage-vector vt)
	 (array-storage-vector rwork))
      (let ((aa (blas-array-alien a))
	    (as (double-array-alien s))
	    (au (blas-array-alien u))
	    (avt (blas-array-alien vt))
	    (arwork (double-array-alien rwork)))
	(flet ((f (work lwork)
		 (with-pinned-objects ((array-storage-vector work))
		   (let ((info
			  (nth-value
			   1
			   (zgesvd
			    "S" "S" m n aa m as au m avt min
			    (blas-array-alien work) lwork arwork))))
		     (unless (zerop info)
		       (if (< info 0)
			   (error "Illegal value of parameter ~A in ZGESVD."
				  (- info))
			   (error "ZGESVD did not converge (INFO = ~A)."
				  info)))))))
	  (let ((work (make-blas-array 1)))
	    (f work -1)
	    (let ((lwork (floor (realpart (aref work 0)))))
	      (f (make-blas-array lwork) lwork))))))))

(defun array-decompose-zgesdd (m n a s u vt)
  (let* ((min (min m n))
	 (max (max m n))
	 (rwork (make-double-array
		 (* min (max (+ (* 5 min) 5) (+ (* 2 max) (* 2 min) 1)))))
	 (iwork (make-integer-array (* 8 min))))
    (with-pinned-objects
	((array-storage-vector a)
	 (array-storage-vector s)
	 (array-storage-vector u)
	 (array-storage-vector vt)
	 (array-storage-vector rwork)
	 (array-storage-vector iwork))
      (let ((aa (blas-array-alien a))
	    (as (double-array-alien s))
	    (au (blas-array-alien u))
	    (avt (blas-array-alien vt))
	    (arwork (double-array-alien rwork))
	    (aiwork (integer-array-alien iwork)))
	(flet ((f (work lwork)
		 (with-pinned-objects ((array-storage-vector work))
		   (let ((info
			  (nth-value
			   1
			   (zgesdd
			    "S" m n aa m as au m avt min
			    (blas-array-alien work) lwork arwork aiwork))))
		     (unless (zerop info)
		       (if (< info 0)
			   (error "Illegal value of parameter ~A in ZGESDD."
				  (- info))
			   (error "ZGESDD did not converge (INFO = ~A)."
				  info)))))))
	  (let ((work (make-blas-array 1)))
	    (f work -1)
	    (let ((lwork (floor (realpart (aref work 0)))))
	      (f (make-blas-array lwork) lwork))))))))

(defun array-decompose-zgesvdx (m n a s u vt)
  (let* ((min (min m n))
	 (rwork (make-double-array (* 17 min min)))
	 (iwork (make-integer-array (* 12 min))))
    (with-pinned-objects
	((array-storage-vector a)
	 (array-storage-vector s)
	 (array-storage-vector u)
	 (array-storage-vector vt)
	 (array-storage-vector rwork)
	 (array-storage-vector iwork))
      (let ((aa (blas-array-alien a))
	    (as (double-array-alien s))
	    (au (blas-array-alien u))
	    (avt (blas-array-alien vt))
	    (arwork (double-array-alien rwork))
	    (aiwork (integer-array-alien iwork)))
	(flet ((f (work lwork)
		 (with-pinned-objects ((array-storage-vector work))
		   (let ((info
			  (nth-value
			   2
			   (zgesvdx
			    "V" "V" "A" m n aa m 0d0 0d0 0 0 as au m avt min
			    (blas-array-alien work) lwork arwork aiwork))))
		     (unless (zerop info)
		       (if (< info 0)
			   (error
			    "Illegal value of parameter ~A in ZGESVDX."
			    (- info))
			   (if (eql info (1+ (* 2 n)))
			       (error
				"Internal error occured in DBDSVDX/ZGESVDX.")
			       (error
				"ZGESVDX did not converge (INFO = ~A)."
				info))))))))
	  (let ((work (make-blas-array 1)))
	    (f work -1)
	    (let ((lwork (floor (realpart (aref work 0)))))
	      (f (make-blas-array lwork) lwork))))))))

(defun array-decompose-fallback (m n a s u vt)
  (handler-case (array-decompose-zgesdd m n (array-copy a) s u vt)
    (error (condition)
      (format *error-output* "ZGESDD error: ~A~%Using ZGESVD.~%" condition)
      (array-decompose-zgesvd m n a s u vt))))

(defvar *svd-driver* :fallback)

(defun array-decompose (array indices)
  (if (numberp indices)
      (setf indices (list indices)))
  ;; don't destroy the original array
  (let* ((a (array-copy array))
	 (r (array-rank a)))

    ;; permute indices if they aren't in the correct order
    (unless (and (eql (indices-position r indices) :left)
		 (apply #'< indices))
      (let ((l (loop for i below (length indices) collect i)))
	(setf a (array-permute a (make-indices-permutation r indices l)))
	(setf indices l)))

    (let* ((d (array-dimensions a))
	   (dl (loop
		  for i in indices
		  collect (nth i d)))
	   (dr (loop
		  for i below r
		  unless (member i indices)
		  collect (nth i d)))
	   (m (apply #'* dr))
	   (n (apply #'* dl))
	   (min (min m n))
	   (s (make-double-array min))
	   (u (make-blas-array (append (list min) dr)))
	   (vt (make-blas-array (append dl (list min)))))

      (funcall
       (ecase *svd-driver*
	 (:zgesvd #'array-decompose-zgesvd)
	 (:zgesdd #'array-decompose-zgesdd)
	 (:zgesvdx #'array-decompose-zgesvdx)
	 (:fallback #'array-decompose-fallback))
       m n a s u vt)
      (values vt s u))))

(defun array-addf (x y)
  (unless (= (array-total-size x) (array-total-size y))
    (error "Array dimensions do not match."))
  (with-pinned-objects
      ((array-storage-vector x)
       (array-storage-vector y)
       (array-storage-vector *blas-alien-1*))
    (zaxpy (array-total-size x) (blas-array-alien *blas-alien-1*)
	   (blas-array-alien y) 1 (blas-array-alien x) 1)))
