(defpackage :bhm
  (:use :common-lisp :blas :tensor :mps)
  (:export :ket-physical-index
	   :bra-physical-index
	   :bose-hubbard-site-hamiltonian
	   :binary-bose-hubbard-site-hamiltonian))

(in-package :bhm)

(defun ket-physical-index (n)
  (loop for i below n collect
       (make-segment :numbers (list i) :dimension 1)))

(defun bra-physical-index (n)
  (loop for i below n collect
       (make-segment :numbers (list (- i)) :dimension 1)))

;;
;; bose-hubbard site hamiltonian
;;
;;         -1    0 0       1
;;
;;            +------+
;; -1         | b+ 0 |
;;    +-------+------+-------+
;;  0 |     0 |  1 0 |     0 |
;;  0 | -j*b+ |  h 1 | -j*b- |
;;    +-------+------+-------+
;;  1         | b- 0 |
;;            +------+
;;

(defun bose-hubbard-site-operators (j u e)
  (list
   (cons    '0 #'(lambda (n) (declare (ignore n)) 0))
   (cons    '1 #'(lambda (n) (declare (ignore n)) 1))
   (cons   'b- #'(lambda (n) (sqrt n)))
   (cons   'b+ #'(lambda (n) (sqrt (1+ n))))
   (cons '-jb- #'(lambda (n) (* (- j) (sqrt n))))
   (cons '-jb+ #'(lambda (n) (* (- j) (sqrt (1+ n)))))
   (cons    'h #'(lambda (n) (+ (* 1/2 u n (1- n)) (* e n))))))

(defun bose-hubbard-site-hamiltonian (dimension j u e)
  (let ((ops (bose-hubbard-site-operators j u e))
	(tbl (flet ((f (x y z)
		      (cons (list (make-subscript
				   :numbers (list (first x))
				   :subscript (second x))
				  (make-subscript
				   :numbers (list (first y))
				   :subscript (second y)))
			    z)))
	       (list
		(f '(-1 0) '( 0 0)   'b+)
		(f '(-1 0) '( 0 1)    '0)
		(f '( 0 0) '(-1 0)    '0)
		(f '( 0 0) '( 0 0)    '1)
		(f '( 0 0) '( 0 1)     0)
		(f '( 0 0) '( 1 0)    '0)
		(f '( 0 1) '(-1 0) '-jb+)
		(f '( 0 1) '( 0 0)    'h)
		(f '( 0 1) '( 0 1)    '1)
		(f '( 0 1) '( 1 0) '-jb-)
		(f '( 1 0) '( 0 0)   'b-)
		(f '( 1 0) '( 0 1)    '0)))))
    (let ((l (list (make-segment :numbers '(-1) :dimension 1)
		   (make-segment :numbers '(0) :dimension 2)
		   (make-segment :numbers '(1) :dimension 1))))
      (functional-tensor
       (list l (ket-physical-index dimension) (bra-physical-index dimension) l)
       #'(lambda (s)
	(let ((x (find
		  (list (first s) (fourth s))
		  tbl
		  :test #'equalp
		  :key #'car)))
	  (if x
	      (funcall (cdr (find (cdr x) ops :key #'car))
		       (- (first (subscript-numbers (third s)))))
	      0)))))))

;;
;; binary bose-hubbard site hamiltonian
;;

(defun bose-hubbard-site-hamiltonian-a (dimension j u e ux)
  (let ((ops (cons
	      (cons 'uxn #'(lambda (n) (* ux n)))
	      (bose-hubbard-site-operators j u e)))
	(tbl (flet ((f (x y z)
		      (cons (list (make-subscript
				   :numbers (first x)
				   :subscript (second x))
				  (make-subscript
				   :numbers (first y)
				   :subscript (second y)))
			    z)))
	       (list
		(f '((-1  0) 0) '(( 0  0) 0)   'b+)
		(f '(( 0 -1) 0) '(( 0  1) 0)    '1)
		(f '(( 0  0) 0) '(( 0  0) 0)    '1)
		(f '(( 0  0) 1) '((-1  0) 0) '-jb+)
		(f '(( 0  0) 1) '(( 0  0) 0)    'h)
		(f '(( 0  0) 1) '(( 0  0) 1)    'uxn)
		(f '(( 0  0) 1) '(( 0  0) 2)    '1)
		(f '(( 0  0) 1) '(( 1  0) 0) '-jb-)
		(f '(( 0  1) 0) '(( 0 -1) 0)    '1)
		(f '(( 1  0) 0) '(( 0  0) 0)   'b-)))))
    (functional-tensor
     (list
      (list (make-segment :numbers '(-1  0) :dimension 1)
	    (make-segment :numbers '( 0 -1) :dimension 1)
	    (make-segment :numbers '( 0  0) :dimension 2)
	    (make-segment :numbers '( 0  1) :dimension 1)
	    (make-segment :numbers '( 1  0) :dimension 1))
      (loop for i below dimension collect
	   (make-segment :numbers (list i 0) :dimension 1))
      (loop for i below dimension collect
	   (make-segment :numbers (list (- i) 0) :dimension 1))
      (list (make-segment :numbers '(-1  0) :dimension 1)
	    (make-segment :numbers '( 0 -1) :dimension 1)
	    (make-segment :numbers '( 0  0) :dimension 3)
	    (make-segment :numbers '( 0  1) :dimension 1)
	    (make-segment :numbers '( 1  0) :dimension 1)))
     #'(lambda (s)
      (let ((x (find
		(list (first s) (fourth s))
		tbl
		:test #'equalp
		:key #'car)))
	(if x
	    (funcall (cdr (find (cdr x) ops :key #'car))
		     (- (first (subscript-numbers (third s)))))
	    0))))))

(defun bose-hubbard-site-hamiltonian-b (dimension j u e)
  (let ((ops (cons
	      (cons 'n #'(lambda (n) n))
	      (bose-hubbard-site-operators j u e)))
	(tbl (flet ((f (x y z)
		      (cons (list (make-subscript
				   :numbers (first x)
				   :subscript (second x))
				  (make-subscript
				   :numbers (first y)
				   :subscript (second y)))
			    z)))
	       (list
		(f '((-1  0) 0) '(( 1  0) 0)    '1)
		(f '(( 0 -1) 0) '(( 0  0) 0)   'b+)
		(f '(( 0  0) 0) '(( 0  0) 0)    '1)
		(f '(( 0  0) 1) '(( 0  0) 0)    'n)
		(f '(( 0  0) 2) '(( 0 -1) 0) '-jb+)
		(f '(( 0  0) 2) '(( 0  0) 0)    'h)
		(f '(( 0  0) 2) '(( 0  0) 1)    '1)
		(f '(( 0  0) 2) '(( 0  1) 0) '-jb-)
		(f '(( 0  1) 0) '(( 0  0) 0)   'b-)
		(f '(( 1  0) 0) '((-1  0) 0)    '1)))))
    (functional-tensor
     (list
      (list (make-segment :numbers '(-1  0) :dimension 1)
	    (make-segment :numbers '( 0 -1) :dimension 1)
	    (make-segment :numbers '( 0  0) :dimension 3)
	    (make-segment :numbers '( 0  1) :dimension 1)
	    (make-segment :numbers '( 1  0) :dimension 1))
      (loop for i below dimension collect
	   (make-segment :numbers (list 0 i) :dimension 1))
      (loop for i below dimension collect
	   (make-segment :numbers (list 0 (- i)) :dimension 1))
      (list (make-segment :numbers '(-1  0) :dimension 1)
	    (make-segment :numbers '( 0 -1) :dimension 1)
	    (make-segment :numbers '( 0  0) :dimension 2)
	    (make-segment :numbers '( 0  1) :dimension 1)
	    (make-segment :numbers '( 1  0) :dimension 1)))
     #'(lambda (s)
      (let ((x (find
		(list (first s) (fourth s))
		tbl
		:test #'equalp
		:key #'car)))
	(if x
	    (funcall (cdr (find (cdr x) ops :key #'car))
		     (- (second (subscript-numbers (third s)))))
	    0))))))

(defun binary-bose-hubbard-site-hamiltonian (dimensiona dimensionb ja jb ua ub ux ea eb)
  (let ((a (bose-hubbard-site-hamiltonian-a dimensiona ja ua ea ux))
	(b (bose-hubbard-site-hamiltonian-b dimensionb jb ub eb)))
    (tensor-fuse
     (tensor-permute
      (tensor-contract a 3 b 0)
      '(0 1 3 2 4 5))
     '(1 3 5)
     '(:normal :normal :reverse :reverse))))
