(defpackage :exp
  (:use :common-lisp :sb-ext :blas :array)
  (:export :hermitian-matrix-exponential))

(in-package :exp)

(defun hermitian-matrix-eigenvalues (matrix)
  ;; do not destroy the original matrix
  (setf matrix (array-copy matrix))
  (unless (and (= (array-rank matrix) 2)
	       (= (array-dimension matrix 0) (array-dimension matrix 1)))
    (error "Matrix has incorrect dimensions."))
  (let* ((n (array-dimension matrix 0))
	 (abstol (dlamch "s"))
	 (w (make-double-array n))
	 (z (make-blas-array (list n n)))
	 (isuppz (make-integer-array (* 2 n)))
	 (work (make-blas-array 1))
	 (lwork -1)
	 (rwork (make-double-array 1))
	 (lrwork -1)
	 (iwork (make-integer-array 1))
	 (liwork -1))
    (sb-sys:with-pinned-objects
	((array-storage-vector matrix)
	 (array-storage-vector w)
	 (array-storage-vector z)
	 (array-storage-vector isuppz))
      (let ((ba (double-array-alien matrix))
	    (bw (double-array-alien w))
	    (bz (double-array-alien z))
	    (bisuppz (integer-array-alien isuppz)))
	(macrolet ((m ()
		     ;; todo: zheevr returns info parameter, check it
		     `(sb-sys:with-pinned-objects
			  ((array-storage-vector work)
			   (array-storage-vector rwork)
			   (array-storage-vector iwork))
			(let ((bwork (double-array-alien work))
			      (brwork (double-array-alien rwork))
			      (biwork (integer-array-alien iwork)))
			  (zheevr "v" "a" "u" n ba n 0d0 0d0 0 0 abstol 0 bw bz n
				  bisuppz bwork lwork brwork lrwork biwork liwork)))))
	  (m)
	  (setf lwork (floor (realpart (aref work 0))))
	  (setf work (make-blas-array lwork))
	  (setf lrwork (floor (aref rwork 0)))
	  (setf rwork (make-double-array lrwork))
	  (setf liwork (aref iwork 0))
	  (setf iwork (make-integer-array liwork))
	  (m))))
    ;; m = v*e*v'; return (e v')
    (values w z)))

;; diagonal-matrix-* is a destructive operation

(defun diagonal-matrix-* (diagonal matrix)
  (let ((m (array-dimension matrix 0))
        (n (array-dimension matrix 1)))
    (dotimes (i m matrix)
      (let ((x (aref diagonal i)))
        (dotimes (j n)
          (setf (aref matrix i j) (* (aref matrix i j) x)))))))

(defun hermitian-matrix-exponential (matrix &optional (parameter 1))
  (multiple-value-bind (e w) (hermitian-matrix-eigenvalues matrix)
    (let* ((n (array-dimension e 0))
	   (f (make-blas-array n))
	   (v (array-conjugate w)))
      (dotimes (i n)
	(setf (aref f i)
	      (coerce (exp (* parameter (aref e i)))
		      '(complex double-float))))
      (diagonal-matrix-* f w)
      (array-contract v 0 w 0))))
