(defpackage :mpi
  (:use :common-lisp :sb-alien :serial)
  (:export :mpi-comm
	   :mpi-datatype
	   :mpi-request
	   :mpi-status
	   :*mpi-comm-world*
	   :*mpi-byte*
	   :*mpi-status-ignore*
	   :+mpi-any-source+
	   :alien-mpi-init
	   :alien-mpi-finalize
	   :alien-mpi-comm-size
	   :alien-mpi-comm-rank
	   :alien-mpi-get-processor-name
	   :alien-mpi-send
	   :alien-mpi-ssend
	   :alien-mpi-issend
	   :alien-mpi-wait
	   :alien-mpi-recv
	   :alien-mpi-probe
	   :alien-mpi-get-count
	   :mpi-init
	   :mpi-finalize
	   :mpi-comm-size
	   :mpi-comm-rank
	   :mpi-get-processor-name
	   :mpi-send-object
	   :mpi-issend-object
	   :mpi-wait
	   :mpi-receive-object
	   :mpi-wtime))

(in-package :mpi)

#+openmpi
(progn
  (define-alien-type nil (struct ompi-communicator-t))
  (define-alien-type nil (struct ompi-datatype-t))
  (define-alien-type nil (struct ompi-request-t))
  (define-alien-type nil (struct ompi-status-t
				 (mpi-source int)
				 (mpi-tag int)
				 (mpi-error int)
				 (-count int)
				 (-cancelled int)))

  (define-alien-type mpi-comm (* (struct ompi-communicator-t)))
  (define-alien-type mpi-datatype (* (struct ompi-datatype-t)))
  (define-alien-type mpi-request (* (struct ompi-request-t)))
  (define-alien-type mpi-status (struct ompi-status-t))

  (define-alien-variable "ompi_mpi_comm_world" (struct ompi-communicator-t))
  (define-alien-variable "ompi_mpi_byte" (struct ompi-datatype-t))

  (defparameter *mpi-comm-world* (addr ompi-mpi-comm-world))
  (defparameter *mpi-byte* (addr ompi-mpi-byte))
  (defparameter *mpi-status-ignore* (sb-sys:int-sap 0))
  (defconstant +mpi-any-source+ -1))

#+mpich
(progn
  (define-alien-type mpi-comm int)
  (define-alien-type mpi-datatype int)
  (define-alien-type mpi-request int)
  (define-alien-type mpi-status (struct mpi-status
					(count int)
					(cancelled int)
					(mpi-source int)
					(mpi-tag int)
					(mpi-error int)))

  (defparameter *mpi-comm-world* #x44000000)
  (defparameter *mpi-byte* #x4c00010d)
  (defparameter *mpi-status-ignore* (sb-sys:int-sap 1))
  (defconstant +mpi-any-source+ -2))

(declaim (inline alien-mpi-init))
(define-alien-routine ("MPI_Init" alien-mpi-init) int
  (argc (* int))
  (argv (* (array c-string))))

(declaim (inline alien-mpi-finalize))
(define-alien-routine ("MPI_Finalize" alien-mpi-finalize) int)

(declaim (inline alien-mpi-comm-size))
(define-alien-routine ("MPI_Comm_size" alien-mpi-comm-size) int
  (comm mpi-comm)
  (size (* int)))

(declaim (inline alien-mpi-comm-rank))
(define-alien-routine ("MPI_Comm_rank" alien-mpi-comm-rank) int
  (comm mpi-comm)
  (rank (* int)))

(defconstant +mpi-max-processor-name+ 256)

(declaim (inline alien-mpi-get-processor-name))
(define-alien-routine ("MPI_Get_processor_name" alien-mpi-get-processor-name) int
  (name (* char))
  (resultlen (* int)))

(declaim (inline alien-mpi-send))
(define-alien-routine ("MPI_Send" alien-mpi-send) int
  (buf (* t))
  (count int)
  (datatype mpi-datatype)
  (dest int)
  (tag int)
  (comm mpi-comm))

(declaim (inline alien-mpi-ssend))
(define-alien-routine ("MPI_Ssend" alien-mpi-ssend) int
  (buf (* t))
  (count int)
  (datatype mpi-datatype)
  (dest int)
  (tag int)
  (comm mpi-comm))

(declaim (inline alien-mpi-issend))
(define-alien-routine ("MPI_Issend" alien-mpi-issend) int
  (buf (* t))
  (count int)
  (datatype mpi-datatype)
  (dest int)
  (tag int)
  (comm mpi-comm)
  (request (* mpi-request)))

(declaim (inline alien-mpi-wait))
(define-alien-routine ("MPI_Wait" alien-mpi-wait) int
  (request (* mpi-request))
  (status (* mpi-status)))

(declaim (inline alien-mpi-recv))
(define-alien-routine ("MPI_Recv" alien-mpi-recv) int
  (buf (* t))
  (count int)
  (datatype mpi-datatype)
  (source int)
  (tag int)
  (comm mpi-comm)
  (status (* mpi-status)))

(declaim (inline alien-mpi-probe))
(define-alien-routine ("MPI_Probe" alien-mpi-probe) int
  (source int)
  (tag int)
  (comm mpi-comm)
  (status (* mpi-status)))

(declaim (inline alien-mpi-get-count))
(define-alien-routine ("MPI_Get_count" alien-mpi-get-count) int
  (status (* mpi-status))
  (datatype mpi-datatype)
  (count (* int)))

(declaim (inline alien-mpi-wtime))
(define-alien-routine ("MPI_Wtime" alien-mpi-wtime) double)

(defun mpi-init ()
  (alien-mpi-init nil nil))

(defun mpi-finalize ()
  (alien-mpi-finalize))

(defun mpi-comm-size (comm)
  (with-alien ((size int))
    (alien-mpi-comm-size comm (addr size))
    size))

(defun mpi-comm-rank (comm)
  (with-alien ((rank int))
    (alien-mpi-comm-rank comm (addr rank))
    rank))

(defun mpi-get-processor-name ()
  (with-alien ((name (* char) (make-alien char +mpi-max-processor-name+))
	       (resultlen int))
    (prog2
	(alien-mpi-get-processor-name name (addr resultlen))
	(cast name c-string)
      (free-alien name))))

(defun mpi-send-object (object dest)
  (let ((buf (serialize object)))
    (sb-sys:with-pinned-objects
	((sb-ext:array-storage-vector buf))
      (alien-mpi-send
       (sap-alien
	(sb-sys:vector-sap (sb-ext:array-storage-vector buf))
	(* char))
       (length buf)
       *mpi-byte*
       dest
       0
       *mpi-comm-world*))))

;; external without-gcing

(defun mpi-issend-object (object dest)
  (let ((buf (serialize object))
	(request (make-alien mpi-request)))
    (alien-mpi-issend
     (sap-alien
      (sb-sys:vector-sap (sb-ext:array-storage-vector buf))
      (* char))
     (length buf)
     *mpi-byte*
     dest
     0
     *mpi-comm-world*
     request)
    request))

(defun mpi-wait (request)
  (alien-mpi-wait
   request
   *mpi-status-ignore*)
  (free-alien request))

(defun mpi-receive-object (source)
  (with-alien ((status mpi-status)
	       (count int))
    (alien-mpi-probe source 0 *mpi-comm-world* (addr status))
    (alien-mpi-get-count (addr status) *mpi-byte* (addr count))
    (let ((buf (make-array count :element-type '(unsigned-byte 8))))
      (sb-sys:with-pinned-objects
	  ((sb-ext:array-storage-vector buf))
	(alien-mpi-recv
	 (sap-alien
	  (sb-sys:vector-sap (sb-ext:array-storage-vector buf))
	  (* char))
	 count
	 *mpi-byte*
	 (slot status 'mpi-source)
	 (slot status 'mpi-tag)
	 *mpi-comm-world*
	 *mpi-status-ignore*))
      (values (deserialize buf) (slot status 'mpi-source)))))

(defun mpi-wtime ()
  (alien-mpi-wtime))
