(defpackage :mps
  (:use :common-lisp :blas :mpi :tensor)
  (:export :make-submps
	   :submps-start
	   :submps-size
	   :submps-tensors
	   :submps-singvals
	   :left-open-boundary-tensor
	   :right-open-boundary-tensor
	   :mps-prepare
	   :mps-overlap
	   :mps-expectation-value
	   :save-metadata
	   :save-submps
	   :save-mps
	   :state-parameter
	   :state-tensor
	   :load-mps))

(in-package :mps)

(defstruct submps
  start
  size
  tensors
  singvals)

(defun left-open-boundary-tensor (operator)
  (let* ((i (first (tensor-indices operator)))
	 (d (segment-dimension (find-if #'list-zerop i :key #'segment-numbers))))
    (functional-tensor
     (list (list (make-segment :numbers '(0) :dimension 1))
	   (conjugate-index i))
     #'(lambda (s)
      (let ((u (subscript-subscript (second s))))
	(if (= u (1- d))
	    1
	    0))))))

(defun right-open-boundary-tensor (operator)
  (let* ((i (fourth (tensor-indices operator))))
    (functional-tensor
     (list (conjugate-index i)
	   (list (make-segment :numbers '(0) :dimension 1)))
     #'(lambda (s)
      (if (zerop (subscript-subscript (first s)))
	  1
	  0)))))

(defun right-singvals-tensor (index)
  (functional-tensor
   (list
    (mapcar #'(lambda (s)
		(make-segment :numbers (list-- (segment-numbers s))
			      :dimension (segment-dimension s)))
	    index)
    index)
   #'(lambda (s) (declare (ignore s)) 1)))

(defun mps-prepare (rank length submps maxdim)
  (let* ((s (submps-start submps))
	 (d (submps-size submps))
	 (u (submps-tensors submps))
	 (v (make-array (length u)))
	 l w r)
    (when (< s length)
      (setf r (if (zerop s)
		  (let ((m (list (make-segment :numbers '(0) :dimension 1))))
		    (functional-tensor
		     (list m m)
		     #'(lambda (s) (declare (ignore s)) 1)))
		  (mpi-receive-object (1- rank))))
      (loop for i from 0 below d do
	   (setf (svref u i) (tensor-contract r 1 (svref u i) 0))
	   (multiple-value-setq (l w r) (tensor-decompose (svref u i) '(0 1) maxdim :left))
	   (setf (svref u i) l)
	   (setf (svref v i) w))
      (if (= (+ s d) length)
	  (setf (svref v (1- d)) (right-singvals-tensor (nth 2 (tensor-indices (svref u (1- d))))))
	  (mpi-send-object r (1+ rank))))
    (setf (submps-singvals submps) v)))

(defun mps-overlap (size rank length submps1 submps2)
  (let ((s (submps-start submps1))
	(d (submps-size submps1))
	(u1 (submps-tensors submps1))
	(u2 (submps-tensors submps2))
	l r)
    (setf l (if (zerop s)
		(let ((m (list (make-segment :numbers '(0) :dimension 1))))
		  (functional-tensor (list m m)
				     #'(lambda (s) (declare (ignore s)) 1)))
		(mpi-receive-object (1- rank))))
    (loop for i from 0 below d do
	 (setf l (tensor-contract
		  (tensor-conjugate (svref u2 i))
		  '(0 1)
		  (tensor-contract l 1 (svref u1 i) 0)
		  '(0 1))))
    (if (and (not (zerop d))
	     (= (+ s d) length))
	(progn
	  (setf r (functional-tensor
		   (list (nth 2 (tensor-indices (svref u2 (1- d))))
			 (conjugate-index (nth 2 (tensor-indices (svref u1 (1- d))))))
		   #'(lambda (s) (declare (ignore s)) 1)))
	  (setf l (tensor-contract l '(0 1) r '(0 1)))))
    (if (= rank (1- size))
	(tensor-scalar l)
	(mpi-send-object l (1+ rank)))))

(defun mps-expectation-value (size rank length mpo submps)
  (let ((s (submps-start submps))
	(d (submps-size submps))
	(u (submps-tensors submps))
	l r)
    (setf l (if (zerop s)
		(let ((m (list (make-segment :numbers '(0) :dimension 1))))
		  (functional-tensor (list m m m)
				     #'(lambda (s) (declare (ignore s)) 1)))
		(mpi-receive-object (1- rank))))
    (loop
       for i from 0 below d
       for j from s
       do
	 (setf l (tensor-contract
		  (tensor-conjugate (svref u i))
		  '(0 1)
		  (tensor-contract (svref mpo j) '(0 2)
				   (tensor-contract l 2 (svref u i) 0) '(1 2))
		  '(2 0))))
    (if (and (not (zerop d))
	     (= (+ s d) length))
	(progn
	  (setf r (functional-tensor
		   (list (nth 2 (tensor-indices (svref u (1- d))))
			 (list (make-segment :numbers '(0) :dimension 1))
			 (conjugate-index (nth 2 (tensor-indices (svref u (1- d))))))
		   #'(lambda (s) (declare (ignore s)) 1)))
	  (setf l (tensor-contract l '(0 1 2) r '(0 1 2)))))
    (if (= rank (1- size))
	(tensor-scalar l)
	(mpi-send-object l (1+ rank)))))

(defun save-metadata (dimension length particles pathname)
  (with-open-file (s (ensure-directories-exist (merge-pathnames #p"metadata" pathname))
		     :direction :output
		     :if-exists :supersede)
    (prin1 (list
	    (cons :site-dimension dimension)
	    (cons :model-length length)
	    (cons :particle-number particles))
	   s)
    (terpri s)))

(defun save-tensor (tensor pathname index)
  (with-open-file (s (merge-pathnames (pathname (format nil "tensor.~A" index)) pathname)
		     :direction :output
		     :if-exists :supersede)
    (prin1 tensor s)
    (terpri s)))

(defun save-submps (submps pathname)
  (with-standard-io-syntax
    (loop
       repeat (submps-size submps)
       for i from (submps-start submps)
       for a across (submps-tensors submps)
       do (save-tensor a pathname i))))

(defun save-mps (mps dimension particles pathname)
  (with-standard-io-syntax
    (let ((*print-pretty* t))
      (when (or (not (probe-file pathname)) (y-or-n-p "Overwrite state ~S?" pathname))
	(save-metadata dimension (length mps) particles pathname)
	(loop
	   for i below (length mps)
	   do (save-tensor (svref mps i) pathname i))))))

(defun state-parameter (pathname parameter)
  (cdr (assoc parameter
	      (with-open-file (s (merge-pathnames #p"metadata" pathname))
		(read s)))))

(defun state-tensor (pathname index)
  (with-standard-io-syntax
    (with-open-file (s (merge-pathnames (pathname (format nil "tensor.~A" index)) pathname))
      (read s))))

(defun load-submps (pathname start end)
  (loop
     with d = (- end start)
     with a = (make-array (1+ d))
     for i from 0
     for j from start below end
     do (setf (svref a i) (state-tensor pathname j))
     finally (return (make-submps :start start :size d :tensors a))))

(defun load-mps (rank partition pathname)
  (let ((s (if (zerop rank) 0 (nth (1- rank) partition)))
	(e (nth rank partition)))
    (load-submps pathname s e)))
