(defpackage :part
  (:use :common-lisp :mpi :mps)
  (:export :uniform-partition
	   :balance-partition
	   :change-partition))

(in-package :part)

(defun uniform-partition (size length)
  (multiple-value-bind (q r) (truncate length size)
    (loop
       with d = (max 2 q)
       with e = (if (< q 2) 0 r)
       for i below size
       for j from d by d
       when (or (< (* 2 i) e)
		(<= (* 2 (- size i)) e))
       do (incf j)
       if (< j length)
       collect j
       else
       collect length)))

(defun balanced-partition (size loads)
  (loop
     with l = (mapcar #'rationalize loads)
     with a = (/ (apply #'+ l) size)
     with i = 0
     for c in l
     count c into j
     sum c into s
     when (>= s a)
     if (and (> (- s a) (/ c 2))
	     (> j (+ i 2)))
     collect (1- j) into p
     and
     do
       (decf s a)
       (setf i (1- j))
     else
     when (> j (1+ i))
     collect j into p
     and
     do
       (decf s a)
       (setf i j)
     finally
       (return
	 (let ((q (butlast p)))
	   (nconc
	    q
	    (make-list (- size (length q)) :initial-element (length l)))))))

(defun send-subloads (subloads)
  (mpi-send-object subloads 0))

(defun collect-subloads (size subloads)
  (loop
     with a = (make-array size)
     repeat (1- size)
     do
       (multiple-value-bind (s i) (mpi-receive-object +mpi-any-source+)
	 (setf (svref a i) s))
     finally
       (setf (svref a 0) subloads)
       (return
	 (loop
	    for s across a
	    nconc s))))

(defun balance-partition (size rank subloads)
  (if (zerop rank)
      (let* ((l (collect-subloads size subloads))
	     (p (balanced-partition size l)))
	p)
      (send-subloads subloads)))

(defun instructions (length oldpartition newpartition)
  (loop
     with op = 0
     with np = 0
     for i below length
     when (eql i (car oldpartition))
     do
       (incf op)
       (setf oldpartition (cdr oldpartition))
     when (eql i (car newpartition))
     do
       (incf np)
       (setf newpartition (cdr newpartition))
     unless (= op np)
     collect (list i op np)))

(defun sort-instructions (instructions)
  (loop
     for i in instructions
     when (evenp (cadr i))
     collect i into e
     else
     collect i into o
     finally (return (nconc e o))))

(defun send-instructions (size instructions)
  (let ((a (make-array size :initial-element nil)))
    (dolist (i instructions)
      (push i (svref a (cadr i)))
      (push i (svref a (caddr i))))
    (sb-sys:without-gcing
      (loop
	 for i from 1 below size
	 collect (mpi-issend-object (svref a i) i)
	 into l
	 finally (mapc #'mpi-wait l)))
    (svref a 0)))

(defun receive-instructions ()
  (mpi-receive-object 0))

(defun process-instructions (rank length submps instructions)
  (if instructions
      (let ((l (loop
		  repeat (submps-size submps)
		  for i from (submps-start submps)
		  collect i))
	    s r)
	(loop
	   for i in instructions
	   if (= (cadr i) rank)
	   do
	     (setf l (delete (car i) l))
	     (push i s)
	   else
	   do
	     (push (car i) l)
	     (push i r))
	(let* ((os (submps-start submps))
	       (ou (submps-tensors submps))
	       (ov (submps-singvals submps))
	       (ns (reduce #'min l :initial-value length))
	       (nd (length l))
	       (nu (make-array (1+ nd) :initial-element nil))
	       (nv (make-array (1+ nd) :initial-element nil))
	       m)
	  (sb-sys:without-gcing
	    (dolist (i (nreverse s))
	      (let ((j (- (car i) os))
		    (p (caddr i)))
		(push (mpi-issend-object (svref ou j) p) m)
		(push (mpi-issend-object (svref ov j) p) m)))
	    (dolist (i (nreverse r))
	      (let ((j (- (car i) ns))
		    (p (cadr i)))
		(setf (svref nu j) (mpi-receive-object p))
		(setf (svref nv j) (mpi-receive-object p))))
	    (mapc #'mpi-wait m))
	  (dotimes (i nd)
	    (unless (svref nu i)
	      (let ((j (+ i (- ns os))))
		(setf (svref nu i) (svref ou j))
		(setf (svref nv i) (svref ov j)))))
	  (make-submps :start ns :size nd :tensors nu :singvals nv)))
      submps))

(defun change-partition (size rank length submps oldpartition newpartition)
  (if (zerop rank)
      (let* ((i (instructions length oldpartition newpartition))
	     (s (sort-instructions i))
	     (l (send-instructions size s)))
	(process-instructions rank length submps l))
      (let ((l (receive-instructions)))
	(process-instructions rank length submps l))))
