(defpackage :serial
  (:use :common-lisp :blas :tensor)
  (:export :serialize
	   :deserialize))

(in-package :serial)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defparameter *types*
    '(fixnum double-float list (array blas-float) array sector segment tensor)))

(defparameter *serializers* (make-array (length *types*) :element-type 'compiled-function))
(defparameter *deserializers* (make-array (length *types*) :element-type 'compiled-function))

(defconstant +index-size+ 1)
(defconstant +fixnum-size+ 4)

(defmacro defser (type (obj buf pos) &body body)
  (let ((i (position type *types* :test #'subtypep)))
    `(setf (svref *serializers* ,i)
	   (lambda (,obj ,buf ,pos)
	     (declare
	      (type (array (unsigned-byte 8)) buf)
	      (type fixnum pos))
	     (if (plusp (length buf))
		 (pack ,i +index-size+ ,buf ,pos))
	     (incf pos +index-size+)
	     (+ +index-size+ (progn ,@body))))))

(defmacro defdes (type (buf pos) &body body)
  (let ((i (position type *types* :test #'subtypep)))
    `(setf (svref *deserializers* ,i)
	   (lambda (,buf ,pos)
	     (declare
	      (type fixnum pos))
	     ,@body))))

(declaim (inline pack unpack))

(defun pack (val size buf pos)
  (declare
   (type fixnum val size pos)
   (type (simple-array (unsigned-byte 8)) buf))
  (case size
    (1 (setf
	(sb-sys:sap-ref-8
	 (sb-sys:vector-sap (sb-ext:array-storage-vector buf))
	 pos)
	val))
    (4 (setf
	(sb-sys:signed-sap-ref-32
	 (sb-sys:vector-sap (sb-ext:array-storage-vector buf))
	 pos)
	val))))

(defun unpack (size buf pos)
  (declare
   (type fixnum size pos)
   (type (simple-array (unsigned-byte 8)) buf))
  (case size
    (1 (sb-sys:sap-ref-8
	(sb-sys:vector-sap (sb-ext:array-storage-vector buf))
	pos))
    (4 (sb-sys:signed-sap-ref-32
	(sb-sys:vector-sap (sb-ext:array-storage-vector buf))
	pos))))

(defun ser (obj buf pos)
  (funcall
   (svref
    *serializers*
    #.`(etypecase obj
	 ,@(loop
	      for i below (length *types*)
	      collect (list (nth i *types*) i))))
   obj
   buf
   pos))

(defun des (buf pos)
  (declare (type fixnum pos))
  (multiple-value-bind (val size)
      (funcall
       (svref *deserializers* (unpack +index-size+ buf pos))
       buf
       (+ pos +index-size+))
    (declare (type fixnum size))
    (values val (+ +index-size+ size))))

(defser fixnum (obj buf pos)
  (if (plusp (length buf))
      (pack obj +fixnum-size+ buf pos))
  +fixnum-size+)

(defdes fixnum (buf pos)
  (values (unpack +fixnum-size+ buf pos) +fixnum-size+))

(defser double-float (obj buf pos)
  (if (plusp (length buf))
      (setf
       (sb-sys:sap-ref-double
	(sb-sys:vector-sap (sb-ext:array-storage-vector buf))
	pos)
       obj))
  8)

(defdes double-float (buf pos)
  (values (sb-sys:sap-ref-double
	   (sb-sys:vector-sap (sb-ext:array-storage-vector buf))
	   pos)
	  8))

(defser list (obj buf pos)
  (if (plusp (length buf))
      (pack (length obj) +fixnum-size+ buf pos))
  (let ((i +fixnum-size+))
    (declare (type fixnum i))
    (dolist (x obj i)
      (incf i (the fixnum (ser x buf (+ pos i)))))))

(defdes list (buf pos)
  (let ((n (unpack +fixnum-size+ buf pos))
	(i +fixnum-size+)
	(obj nil))
    (declare (type fixnum n i))
    (dotimes (j n (values (nreverse obj) i))
      (multiple-value-bind (val size) (des buf (+ pos i))
	(declare (type fixnum size))
	(incf i size)
	(push val obj)))))

(defser array (obj buf pos)
  (let* ((d (array-dimensions obj))
	 (i (ser d buf pos)))
    (declare (type fixnum i))
    (dotimes (j (array-total-size obj) i)
      (incf i (the fixnum (ser (row-major-aref obj j) buf (+ pos i)))))))

(defdes array (buf pos)
  (multiple-value-bind (d i) (des buf pos)
    (declare (type list d))
    (declare (type fixnum i))
    (let ((obj (make-array d)))
      (dotimes (j (array-total-size obj) (values obj i))
	(multiple-value-bind (val size) (des buf (+ pos i))
          (declare (type fixnum size))
	  (incf i size)
	  (setf (row-major-aref obj j) val))))))

(defser (array blas-float) (obj buf pos)
  (let* ((d (array-dimensions obj))
	 (n (* (array-total-size obj) +blas-float-size+))
	 (i (ser d buf pos)))
    (declare (type fixnum n i))
    (if (plusp (length buf))
	(sb-sys:with-pinned-objects
	    ((sb-ext:array-storage-vector obj)
	     (sb-ext:array-storage-vector buf))
	  (sb-sys:memmove
	   (sb-alien:sap-alien
	    (sb-sys:sap+
	     (sb-sys:vector-sap (sb-ext:array-storage-vector buf))
	     (+ pos i))
	    (* char))
	   (sb-alien:sap-alien
	    (sb-sys:vector-sap (sb-ext:array-storage-vector obj))
	    (* char))
	   n)))
    (+ i n)))

(defdes (array blas-float) (buf pos)
  (multiple-value-bind (d i) (des buf pos)
    (declare (type list d))
    (declare (type fixnum i))
    (let* ((obj (make-blas-array d))
	   (n (* (array-total-size obj) +blas-float-size+)))
      (declare (type fixnum n))
      (sb-sys:with-pinned-objects
	  ((sb-ext:array-storage-vector obj)
	   (sb-ext:array-storage-vector buf))
	(sb-sys:memmove
	 (sb-alien:sap-alien
	  (sb-sys:vector-sap (sb-ext:array-storage-vector obj))
	  (* char))
	 (sb-alien:sap-alien
	  (sb-sys:sap+
	   (sb-sys:vector-sap (sb-ext:array-storage-vector buf))
	   (+ pos i))
	  (* char))
	 n))
      (values obj (+ i n)))))

(defser sector (obj buf pos)
  (let ((i (ser (sector-numbers obj) buf pos)))
    (declare (type fixnum i))
    (+ i (the fixnum (ser (sector-array obj) buf (+ pos i))))))

(defdes sector (buf pos)
  (declare (type fixnum pos))
  (multiple-value-bind (val1 size1) (des buf pos)
    (declare (type fixnum size1))
    (multiple-value-bind (val2 size2) (des buf (+ pos size1))
      (declare (type fixnum size2))
      (values (make-sector :numbers val1 :array val2) (+ size1 size2)))))

(defser segment (obj buf pos)
  (let ((i (ser (segment-numbers obj) buf pos)))
    (declare (type fixnum i))
    (+ i (the fixnum (ser (segment-dimension obj) buf (+ pos i))))))

(defdes segment (buf pos)
  (declare (type fixnum pos))
  (multiple-value-bind (val1 size1) (des buf pos)
    (declare (type fixnum size1))
    (multiple-value-bind (val2 size2) (des buf (+ pos size1))
      (declare (type fixnum size2))
      (values (make-segment :numbers val1 :dimension val2) (+ size1 size2)))))

(defser tensor (obj buf pos)
  (let ((i (ser (tensor-indices obj) buf pos)))
    (declare (type fixnum i))
    (+ i (the fixnum (ser (tensor-sectors obj) buf (+ pos i))))))

(defdes tensor (buf pos)
  (declare (type fixnum pos))
  (multiple-value-bind (val1 size1) (des buf pos)
    (declare (type fixnum size1))
    (multiple-value-bind (val2 size2) (des buf (+ pos size1))
      (declare (type fixnum size2))
      (values (make-tensor :indices val1 :sectors val2) (+ size1 size2)))))

(defun size (obj)
  (ser obj (make-array 0 :element-type '(unsigned-byte 8)) 0))

(defun serialize (obj)
  (let* ((buf (make-array (size obj) :element-type '(unsigned-byte 8))))
    (ser obj buf 0)
    buf))

(defun deserialize (buf)
  (des buf 0))
