(defpackage :tebd
  (:use :common-lisp :blas :tensor :mpi :mps)
  (:export :make-single-site-propagator
	   :make-double-site-propagator
	   :make-mph-propagator
	   :make-uniform-propagator
	   :tebd-evolve))

(in-package :tebd)

(defun left-tensor (operator)
  (let* ((i (first (tensor-indices operator)))
	 (d (segment-dimension (find-if #'list-zerop i :key #'segment-numbers))))
    (functional-tensor
     (list i)
     #'(lambda (s)
      (let ((u (first s)))
	(if (and (list-zerop (subscript-numbers u))
		 (= (subscript-subscript u) (1- d)))
	    1
	    0))))))

(defun right-tensor (operator)
  (let* ((i (fourth (tensor-indices operator))))
    (functional-tensor
     (list i)
     #'(lambda (s)
      (let ((u (first s)))
	(if (and (list-zerop (subscript-numbers u))
		 (= (subscript-subscript u)) 0)
	    1
	    0))))))

(defun central-tensor (operator)
  (let* ((i (first (tensor-indices operator)))
	 (d (segment-dimension (find-if #'list-zerop i :key #'segment-numbers))))
    (functional-tensor
     (list i
	   (reverse
	    (mapcar #'(lambda (s)
			(make-segment :numbers (list-- (segment-numbers s))
				      :dimension (segment-dimension s)))
		    i)))
     #'(lambda (s)
      (let ((u (subscript-subscript (first s)))
	    (v (subscript-subscript (second s))))
	(if (= u v)
	    (if (= u (1- d))
		0
		1)
	    0))))))

(defun make-single-site-propagator (operator step)
  (let* ((l (left-tensor operator))
	 (r (right-tensor operator))
	 (h (tensor-contract l 0 (tensor-contract operator 3 r 0) 0)))
    (tensor-exponential h '((0) (1)) (* -1 +blas-i+ step))))

(defun make-double-site-propagator (operator1 operator2 step)
  (let* ((l (left-tensor operator1))
	 (r (right-tensor operator2))
	 (c (central-tensor operator2))
	 (h (tensor-contract
	     l
	     0
	     (tensor-contract
	      operator1
	      3
	      (tensor-contract c 1 (tensor-contract operator2 3 r 0) 0)
	      0)
	     0)))
    (tensor-exponential h '((0 2) (1 3)) (* -1 +blas-i+ step))))

(defun make-single-propagators (mph step)
  (let ((f (svref mph 0))
	(g (svref mph (1- (length mph)))))
    (list (if f (make-single-site-propagator f step))
	  (if g (make-single-site-propagator g step)))))

(defun make-double-propagators (mph step)
  (loop
     with l = (1- (length mph))
     with a = (make-array l)
     for i below l
     for f = (svref mph i)
     for g = (svref mph (1+ i))
     when (and f g)
     do (setf (svref a i) (make-double-site-propagator f g step))
     finally (return a)))

(defstruct propagator
  single1
  double1
  single2
  double2)

(defun make-mph-propagator (mph step)
  (make-propagator
   :single1 (make-single-propagators mph step)
   :double1 (make-double-propagators mph step)
   :single2 (make-single-propagators mph (/ step 2))
   :double2 (make-double-propagators mph (/ step 2))))

(defun make-uniform-propagator (length hamiltonian step)
  (let ((s1 (make-single-site-propagator hamiltonian step))
	(d1 (make-double-site-propagator hamiltonian hamiltonian step))
	(s2 (make-single-site-propagator hamiltonian (/ step 2)))
	(d2 (make-double-site-propagator hamiltonian hamiltonian (/ step 2))))
    (make-propagator
     :single1 (list s1 s1)
     :double1 (make-array (1- length) :initial-element d1)
     :single2 (list s2 s2)
     :double2 (make-array (1- length) :initial-element d2))))

(defun apply-single-site-operator (operator tensor)
  (tensor-permute (tensor-contract operator 1 tensor 1) '(1 0 2)))

(defun apply-double-site-operator (operator tensor1 tensor2 singvals maxdim)
  (let ((u (tensor-contract
	    operator
	    '(1 3)
	    (tensor-contract tensor1 2 tensor2 0)
	    '(1 2))))
    (multiple-value-bind (l s r)
	(tensor-decompose (tensor-contract u 3 singvals 0) '(2 0) maxdim :left)
      (declare (ignore r))
      (values l s (tensor-contract (tensor-conjugate l) '(0 1) u '(2 0))))))

(defun apply-even-tebd-operators (length single double submps maxdim &key (subwt nil) (gc nil))
  (let* ((s (submps-start submps))
	 (d (submps-size submps))
	 (e (+ s d))
	 (u (submps-tensors submps))
	 (v (submps-singvals submps)))
    (if (and (= e length) (oddp e) (not (zerop d)))
	(let ((w (mpi-wtime)))
	  (setf (svref u (1- d)) (apply-single-site-operator (cadr single) (svref u (1- d))))
	  (when subwt
	    (incf (svref subwt (1- d)) (- (mpi-wtime) w)))))
    (loop
       for i from (if (evenp s) 0 1) below (if (= e length) (1- d) d) by 2
       for w = (mpi-wtime)
       do
	 (when gc
	   (sb-ext:gc :full t))
	 (multiple-value-bind (l s r)
	     (apply-double-site-operator
	      (svref double (+ s i))
	      (svref u i)
	      (svref u (1+ i))
	      (svref v (1+ i))
	      maxdim)
	 (setf (svref u i) l)
	 (setf (svref u (1+ i)) r)
	 (setf (svref v i) s))
	 (when subwt
	   (incf (svref subwt i) (- (mpi-wtime) w))))))

(defun apply-odd-tebd-operators (length single double submps maxdim &key (subwt nil) (gc nil))
  (let* ((s (submps-start submps))
	 (d (submps-size submps))
	 (e (+ s d))
	 (u (submps-tensors submps))
	 (v (submps-singvals submps)))
    (if (zerop s)
	(let ((w (mpi-wtime)))
	  (setf (svref u 0) (apply-single-site-operator (car single) (svref u 0)))
	  (when subwt
	    (incf (svref subwt 0) (- (mpi-wtime) w)))))
    (if (and (= e length) (evenp e) (not (zerop d)))
	(let ((w (mpi-wtime)))
	  (setf (svref u (1- d)) (apply-single-site-operator (cadr single) (svref u (1- d))))
	  (when subwt
	    (incf (svref subwt (1- d)) (- (mpi-wtime) w)))))
    (loop
       for i from (if (evenp s) 1 0) below (if (= e length) (1- d) d) by 2
       for w = (mpi-wtime)
       do
	 (when gc
	   (sb-ext:gc :full t))
	 (multiple-value-bind (l s r)
	     (apply-double-site-operator
	      (svref double (+ s i))
	      (svref u i) (svref u (1+ i))
	      (svref v (1+ i))
	      maxdim)
	 (setf (svref u i) l)
	 (setf (svref u (1+ i)) r)
	 (setf (svref v i) s))
	 (when subwt
	   (incf (svref subwt i) (- (mpi-wtime) w))))))

(defun tebd-exchange-phase-0 (rank length submps)
  (let* ((s (submps-start submps))
	 (d (submps-size submps))
	 (e (+ s d))
	 (u (submps-tensors submps))
	 (v (submps-singvals submps)))
    (when (not (zerop d))
      (let (l)
	(sb-sys:without-gcing
	  (when (and (oddp s) (/= s 0))
	    (push (mpi-issend-object (svref u 0) (1- rank)) l)
	    (push (mpi-issend-object (svref v 0) (1- rank)) l))
	  (when (and (oddp e) (/= e length))
	    (setf (svref u d) (mpi-receive-object (1+ rank)))
	    (setf (svref v d) (mpi-receive-object (1+ rank))))
	  (mapc #'mpi-wait l))))))

(defun tebd-exchange-phase-1 (rank length submps)
  (let* ((s (submps-start submps))
	 (d (submps-size submps))
	 (e (+ s d))
	 (u (submps-tensors submps))
	 (v (submps-singvals submps)))
    (when (not (zerop d))
      (let (l)
	(sb-sys:without-gcing
	  (when (and (oddp e) (/= e length))
	    (push (mpi-issend-object (svref u d) (1+ rank)) l))
	  (when (and (evenp s) (/= s 0))
	    (push (mpi-issend-object (svref u 0) (1- rank)) l)
	    (push (mpi-issend-object (svref v 0) (1- rank)) l))
	  (when (and (oddp s) (/= s 0))
	    (setf (svref u 0) (mpi-receive-object (1- rank))))
	  (when (and (evenp e) (/= e length))
	    (setf (svref u d) (mpi-receive-object (1+ rank)))
	    (setf (svref v d) (mpi-receive-object (1+ rank))))
	  (mapc #'mpi-wait l))))))

(defun tebd-exchange-phase-2 (rank length submps)
  (let* ((s (submps-start submps))
	 (d (submps-size submps))
	 (e (+ s d))
	 (u (submps-tensors submps))
	 (v (submps-singvals submps)))
    (when (not (zerop d))
      (let (l)
	(sb-sys:without-gcing
	  (when (and (evenp e) (/= e length))
	    (push (mpi-issend-object (svref u d) (1+ rank)) l))
	  (when (and (oddp s) (/= s 0))
	    (push (mpi-issend-object (svref u 0) (1- rank)) l)
	    (push (mpi-issend-object (svref v 0) (1- rank)) l))
	  (when (and (evenp s) (/= s 0))
	    (setf (svref u 0) (mpi-receive-object (1- rank))))
	  (when (and (oddp e) (/= e length))
	    (setf (svref u d) (mpi-receive-object (1+ rank)))
	    (setf (svref v d) (mpi-receive-object (1+ rank))))
	  (mapc #'mpi-wait l))))))

(defun tebd-exchange-phase-3 (rank length submps)
  (let* ((s (submps-start submps))
	 (d (submps-size submps))
	 (e (+ s d))
	 (u (submps-tensors submps)))
    (when (not (zerop d))
      (let (l)
	(sb-sys:without-gcing
	  (when (and (oddp e) (/= e length))
	    (push (mpi-issend-object (svref u d) (1+ rank)) l))
	  (when (and (oddp s) (/= s 0))
	    (setf (svref u 0) (mpi-receive-object (1- rank))))
	  (mapc #'mpi-wait l))))))

(defun tebd-evolve (rank length propagator submps maxdim &key (subwt nil) (gc nil))
  (tebd-exchange-phase-0 rank length submps)
  (apply-even-tebd-operators
   length
   (propagator-single2 propagator)
   (propagator-double2 propagator)
   submps
   maxdim
   :subwt subwt
   :gc gc)
  (tebd-exchange-phase-1 rank length submps)
  (apply-odd-tebd-operators
   length
   (propagator-single1 propagator)
   (propagator-double1 propagator)
   submps
   maxdim
   :subwt subwt
   :gc gc)
  (tebd-exchange-phase-2 rank length submps)
  (apply-even-tebd-operators
   length
   (propagator-single2 propagator)
   (propagator-double2 propagator)
   submps
   maxdim
   :subwt subwt
   :gc gc)
  (tebd-exchange-phase-3 rank length submps))
